#ifndef TEXTWINS_HPP
#define TEXTWINS_HPP
///////////////////////////////////////////////////////////////
//                                                           //
//  textwins.hpp - builds text based windows                 //
//  v3.2                                                     //
//                                                           //
///////////////////////////////////////////////////////////////
//                                                           //
//  manual page                                              //
//                                                           //
//  Language:    Borland C, ver 3.1;                         //
//  Platform:    Gateway 386/25, win95                       //
//  Application: CSE687, pr3                                 //
//  Author:      Jim Fawcett, 123 Link Hall, Syracuse Univ   //
//               Syracuse, N.Y., 13244, (315)443-2654        //
//                                                           //
//           copyright (c) 1996, James W. Fawcett            //
//                all rights granted to user                 //
//                    except publication                     //
//                                                           //
///////////////////////////////////////////////////////////////
/*
  win class operations:
  =====================
  win objects manage their own internal screen memory and
  provide operations for reading and writing strings to and
  from screen windows.

  construct and manage windows:
  =============================
    winMgr wind(10);      // declare a list of ten windows
    wm.makeMsgWin();      // make message window
    wm.makeDiaWin();      // make dialog window
    wm.makeMenuWin();     // make menu window
    wm.makeScrollWin();   // make scrolling window
    wind.remove(i);       // remove ith window
    wind.clearDesk();     // clear screen
    wind.paintAll(fc);    // repaint all windows with focus fc [1,0]
    getEvent(ev);         // get keyboard or mouse event, it is global
    wind.processEvent(ev);// dispatch event to window event handlers

  set and get window contents, activate controls:
  ===============================================
    wind[i].size(cols, rows).place(left, top).prompt("a prompt");
    wind[i].paint(foc, invert); // paint window on console screen
    wind[i].putText(char *p);   // send text to window
    wind[i].getText();          // return window text to client
    wind[i].grabText(wind[1]);  // transfer text between windows
    wind[i].process();          // pick item, scroll window, dialog winow...

  get and set window colors:
  ==========================
    wind[i].colors(fg, bg);
    wind[i].FGcolor(color).BGcolor(color);
     int fg = FGcolor();
     int bg = BGcolor();

*/

//


///////////////////////////////////////////////////////////////
//                                                           //
//  maintenance page                                         //
//                                                           //
///////////////////////////////////////////////////////////////
//                                                           //
//  Build Process                                            //
//                                                           //
//  Files Required:                                          //
//    textwins.hpp, textwins.cpp, tmouse.hpp, tmouse.cpp     //
//                                                           //
//  Command Line Compiler Build:                             //
//    tcc -w+ -mh -DTEST_TEXTWINS textwins.cpp, tmouse.cpp   //
//                                                           //
//    note that huge memory model is required if an appli-   //
//    cation will need several windows open simultaneously   //
//                                                           //
//  Integrated Environment Build:                            //
//    - close any open project                               //
//    - set Options/Compiler/Code generation/defines to      //
//      TEST_TEXTWINS                                        //
//    - run                                                  //
//                                                           //
///////////////////////////////////////////////////////////////
/*
    Maintenance History:
    ====================
    ver 3.2 : 24 Sep 96
      - fixed bug in focus reporting at borders
    ver 3.1 : 29 May 96
      - completed mouse support with the help of Jane Yan.
      - changed order of arguments in size(), place() to x, y
    ver 3.0 : 23 May 96
      - (Jane Yan) added a mouse object in text windows
    ver 2.2 : 27 Apr 96
      - added transient window type
      - added copy constructor as a safety precaution.
    ver 2.1 : 17 Apr 96
      - added scrollable windows, keyboard event dispatcher
    ver 2.0 :  7 Apr 96
      - added events, commands, menuWins, and shell for scrollWins
      - decomposed diaWin::paint() into paint() which does not grab
        control and editText() which does, to support mouse control
      - changed text() to getText()
    ver 1.5 : 29 Mar 96
      - fixed bug found by Amalia Garmendia in diaWin::putText(dt).
    ver 1.4 : 23 Mar 96
      - fixed bug in size() by resizing buffer as well as window
        coordinates, and by changing comparison in paint().
    ver 1.3 : 22 Mar 96
      - cosmetic changes to test stub output
    ver 1.2 : 21 Mar 96
      - display text extracted from msgWin and diaWin
    ver 1.1 : 20 Mar 96
      - bug fixes to eliminate buffer overwrite and spurious blinking
    ver 1.0 : 15 Mar 96
      - first release

    Planned Modifications:
    ======================
    - add kill buttons
    - make test stub communicate window expectations more clearly
*/
//


///////////////////////////////////////////////////////////////
//                                                           //
//  design notes - text windows class hierarchy              //
//                                                           //
///////////////////////////////////////////////////////////////
/*

                  +-------+        +--------+      +-------+
                  |  win  |o<----<>| winMgr |<>--->| Event |
                  +---+---+        +--------+      +-------+
                      |
                     / \
     +-----------+------------+-----------------+-------------+
     |           |            |                 |             |
 +---+----+  +---+----+  +----+----+       +----+----+  +-----+-----+
 | msgWin |  | diaWin |  | itemWin |o<---<>| menuWin |  | scrollWin |
 +--------+  +--------+  +---------+       +----+----+  +-----------+
     |                       < >
     |                        |
    / \                       |
 +--------+  +---------+      |
 | trnWin |  | command |o<----+
 +--------+  +----+----+
                  |
                 / \
                  |
             +----+----+
             | testCom |
             +---------+

  classes:
  ========

  win       - base class for all windows.  Provides the ability
              to paint a bordered rectangle on the screen
  msgWin    - adds capability to display a message, centered on
              the window
  trnWin    - makes a message window transient, e.g., hidden on
              the second call to paint(...)
  diaWin    - adds the capability to insert a string from the
              keyboard
  itemWin   - similar to msgWin but with no border
  menuWin   - this parent window has a list of itemWin children.
              When the ith itemWin in its child list is selected
              it invokes the ith item in its command list.
  scrollWin - attaches to file and displays its text in a scroll-
              able window.
  command   - provides an execute() command, called every time a
              an itemWin object is selected from a menuWin.  This
              is an abstract class so a client must define a
              derived class which gives meaning to the execute()
              invocation.  The test stub illustrates this with
              a testCom class.
*/

//


///////////////////////////////////////////////////////////////
//                                                           //
//  design notes - window manager and events                 //
//                                                           //
///////////////////////////////////////////////////////////////
/*


  winMgr class:
  =============

  window manager class manages a list of active text windows
  on the console screen.  window manager has make functions
  for each type of derived window.  Each time a window is made
  it is added to the winMgr's list for subsequent management
  actions.

  event operations:
  =================
    - keys(esc, q, Q, x, X) will terminate program if it is not in
      a process() function for dialog, menu,or scroll window
    - tabs cycle through the windows
    - mouse left button click to select and process window clicked
    - p to process the current active window
    - dialog window: keys(enter, esc, tab) will terminate process;
    - menu window: keys(esc, tab) will terminate process, enter and
      mouse click will invoke select() and terminate process;
    - scroll window: keys(esc, tab, x, X, q) will terminate process;

*/

//


///////////////////////////////////////////////////////////////
//  textwins.hpp - base class declarations                   //
///////////////////////////////////////////////////////////////


#include <conio.h>     // cprintf(), colors
#include "tmouse.hpp"  // tmouse()

const int defTop = 10;                 // default window top row
const int defLft = 20;                 // default window left col
const int defHgt = 10;                 // default window height
const int defWid = 30;                 // default window width
const int defFGC = WHITE;              // default foreground color
const int defBGC = LIGHTBLUE;          // default background color

enum EventType { null, Exitw, Exit, kbChar, mlb, mrb, mm, all, noMotion };
enum   justify { Left, Center };

struct Event {
  EventType type; int chr, ext; int x, y;
};


class win {

  friend class winMgr;

  public:

    win(int lt=defLft, int tp=defTop, int wd=defWid, int ht=defHgt);
    win(const win&);
    virtual ~win();
     win& place(int lt=defLft, int tp=defTop);
     win& prompt(char *p);
      int width();
      int height();
      int top();
      int left();
      int row();
      int col();
     win& colors(int fg, int bg);
     win& FGcolor(int color);
     win& BGcolor(int color);
      int FGcolor();
      int BGcolor();
    const char* getText();
     void index(int i) { myIndex = i; }
      int index() { return myIndex; }
   virtual win& size(int x=defWid, int y=defHgt);
   virtual win& moveText(win &w);
   virtual win& putText(const char *p);
   virtual win& paint(int foc, int invert=0, int just=Center);
   virtual win& process(Event &)=0;
   virtual void alert(const char *m,int row=12,int col=20,int=0);
  static   void alertMsg(const char *m,int row=12,int col=20,int=0);
   virtual  int myEvent(const Event& ev);
   virtual  int handleEvent(Event &)   { return 0; }
  static   void acceptAllMouse() { acceptAll = 1; }
  static   void acceptInWindow() { acceptAll = 0; }
  static Event& getEvent(Event &ev);
  static tmouse mouse;            // single mouse object
//



  protected:

      int lft, tp,  wd, ht;       // left, top, width, height
      int bgc, fgc;               // background, foreground colors
      int *buffer;                // screen buffer holds extended chars
      int bufSize;                // screen buffer size
     char *prtBuf;                // prompt buffer
      int prtSize;                // prompt buffer size
     char *txtBuf;                //   text buffer
      int txtSize;                //   text buffer size
      int inverseVideo;           // reverse foreground and background
      int myIndex;                // order in winMgr's list
   static  int acceptAll;         // accept all mouse positions?
   static  win *alertwin;         // window for error messages

   static char *concat(const char *s1, const char *s2);
     void abort(const char *s);
      int chkHor(int p);
      int chkVer(int p);
     win& charxy(unsigned char ch, int row, int col);
   unsigned int extchar(unsigned char);
     win& horizLine(char ch, int row, int start, int end);
     win& vertiLine(char ch, int col, int start, int end);
     void drawBorders(int intensity);
     void focus(int foc);
     void toggleVideo() { inverseVideo = inverseVideo ? 0 : 1; }
      int textStart(int width, int length);
      int textEnd  (char *buffer, int size);

    virtual  int map(int row, int col);

  private:
    void operator=(const win &) {
      cout << "assignment is not implemented";
    }
};

//


///////////////////////////////////////////////////////////////
//  textwins.hpp - derived class declarations                //
///////////////////////////////////////////////////////////////


class msgWin : public win {     // message window

  public:
    msgWin(int lt=defLft,int tp=defTop,int wd=defWid,int ht=3);
    virtual ~msgWin() { }
    virtual win& process(Event &) { return *this; }
};


class trnWin : public msgWin {  // transient window

  public:
    trnWin(int lt=defLft, int tp=defTop, int wd=defWid, int ht=3);
    virtual ~trnWin();
    virtual win& process(Event &) { return *this; }
    virtual win& size(int x=defWid, int y=defHgt);
    virtual win& paint(int focus ,int invert=0, int justify=0);

  private:
     int *saveBuf;      // save buffer holds obscured background
     int saveSize;      // save buffer size
  static int count;     // only create one
};


class diaWin : public win {     // dialog window

  public:
    diaWin(int lt=defLft,int tp=defTop,int wd=defWid,int ht=3);
    virtual win& putText(const char *p);
    virtual win& process(Event &ev);
    virtual win& paint(int foc, int invert=0, int justify=0);
    virtual ~diaWin() { }

  private:
    int editRow;        // place to start edit
    int editCol;        //   "   "    "    "
   void putChar(char ch, int col, int row);
};

//



class command { public: virtual void execute()=0; };

class itemWin : public win {    // menu item

  public:
    itemWin(int lt=defLft,int tp=defTop,int wd=defWid,int ht=1);
    virtual ~itemWin();
    virtual win& process(Event &) { return *this; }
    void regComm(command *c) { childCom = c; }
    void execute() { childCom->execute(); }

  protected:
    command *childCom;  // command executed when selected

};

class menuWin : public win {    // menu pick list window

  public:
    menuWin(int size, int=defLft,int=defTop,int=defWid,int=5);
    menuWin(const win& w);
    virtual ~menuWin();
    int install(itemWin* item, char *text, command *c);
    virtual win& putText(const char *text);
    virtual win& paint(int foc, int invert=0, int justify=0);
    virtual win& process(Event &ev);
    virtual void select(int n);
    virtual void hiLite(int n);
    virtual void loLite(int n);

  private:
    itemWin** child;
    int next;
    int max;
    int current;     // the hilited item win
   void move(int index);
};


class scrollWin : public win {  // scrolling text window

  public:
    scrollWin::scrollWin(int lt=defLft, int tp=defTop,
                         int wd=defWid, int ht=defHgt);
    virtual ~scrollWin();
    virtual win& putText(const char *file);
    virtual win& paint(int foc, int invert=0, int justify=0);
    virtual win& process(Event &ev);

  private:
    char far **line;
    int lineSize;
    int next;
};


//


///////////////////////////////////////////////////////////////
//  textwins.hpp - window manager class declarations         //
///////////////////////////////////////////////////////////////

enum mgrType { tiled, overlapped };

class winMgr {

  public:
    winMgr(int size, mgrType mt=overlapped);
    virtual ~winMgr();
     int makeDiaWin();
     int makeMsgWin();
     int makeTrnWin();
     int makeMenuWin(menuWin &mw);
     int makeScrollWin();
     int top(int n);
    win& operator[](int n);
    void remove(int n);
    void paintAll(int foc);
    void clearDesk();
   Event waitForEvent(EventType et);
  Event& processEvent(Event &ev);
  static void abort(char *s);

  private:
    mgrType type;     // tiled or overlapped
    win** list;       // window list
    int next;         // next available slot in window list
    int max;          // maximum number of windows in list
    int current;      // active window if tiled
    int scroll;       // conio scroll parameter
    int video;        // inverse video flag
    int tiled;        // don't pop window if tiled
    Event ev;         // for Event
    Event& processKbEvent(Event &ev);
    Event& processMouseEvent(Event &ev);
};

#endif
