///////////////////////////////////////////////////////////////
//                                                           //
//  toks.cpp  -  simple tokenizer returns words, delimiters  //
//  v1.5                                                     //
//               Implementations                             //
//                                                           //
//  Language:    Borland C++, ver 3.1                        //
//  System:      Gateway 386/25, DOS 6.2                     //
//  Application: CSE791 Demonstration                        //
//  Author:      Jim Fawcett, Syracuse University            //
//               123 Link Hall, (315) 443-2654               //
//                                                           //
//                    copyright (c) 1996                     //
//    all rights granted to users except right to publish    //
//                                                           //
///////////////////////////////////////////////////////////////
/*

    Design Notes:
    =============
    - The tokenizer does not return C or C++ comments.
    - Low level statistics are gathered by function getChr
      which reads characters from the input stream.
    - ungetchr() puts a single char back onto the input
      stream, decrementing statistics when that is appropriate
    - Statistics tracked by toks objects are number of lines,
      number of comment lines, number of expressions, and the
      net number of braces.

*/

#include <ctype.h>    // isalnum()
#include "toks.hpp"


//----< void constructor >-------------------------------------

toks::toks(void) : inComm(0), inQuote(0),  numL(0),
                    numCL(0),  numExp(0), numBr(0), in(cin) {

    state = NoExist;
}


//----< constructor >------------------------------------------

toks::toks(istream &input) : inComm(0), inQuote(0),  numL(0),
                              numCL(0),  numExp(0), numBr(0),
                                                   in(input) {

  if(!in.good())
    state = NoExist;
  else
    state = Good;
}

//

//----< constructor >------------------------------------------

ftoks::ftoks(ifstream &input) : toks(input) { }


//----< destructor >-------------------------------------------

ftoks::~ftoks() { if(state != NoExist) ((ifstream&)in).close(); }


//---< get character from input stream >-----------------------

int toks::getchr() {

  char save = in.get();
  switch(save) {
    case '\n' : numL++;
                if( inComm) numCL++;  break;
    case ';'  : if((!inComm)&&(!inQuote)) numExp++; break;
    case '{'  : if((!inComm)&&(!inQuote)) numBr++;  break;
    case '}'  : if((!inComm)&&(!inQuote)) numBr--;  break;
  }
  return save;
}

//---< put character back onto input stream >------------------
//
//  if successful returns char pushed back
//  else returns EOF
//

int toks::ungetchr(char ch) {

  switch(ch) {
    case '\n' : numL--; if(inComm) numCL--; break;
    case ';'  : if((!inComm)&&(!inQuote)) numExp--; break;
    case '{'  : if((!inComm)&&(!inQuote)) numBr--;  break;
    case '}'  : if((!inComm)&&(!inQuote)) numBr++;  break;
  }
  in.putback(ch);
  return in.good() ? ch : EOF;
}

//

//----< eat single quote >-------------------------------------
//
//  Processing:
//    - extracts, discards characters inside a C/C++ single quote.
//  Inputs:
//    - token is buffer which contains quote on return
//  Ouputs:
//    - returns 1 if no EOF, otherwise 0
//

int toks::eatSquote(str &token) {
  char ch, esc=0;
  inQuote = 1;
  while((ch = getchr()) != EOF) {
    token.append(ch);
    if ((ch == '\'')&&(!esc)) {  // non-escaped single quote
      inQuote = 0; return 1;
    }
    if ((ch == '\\')&&(!esc))    // non-escaped escape char
      esc = 1;
    else
      esc = 0;
  }
  inQuote = 0;
  return 0;
}

//----< eat double quote >-------------------------------------
//
//  Processing:
//    - extracts, discards characters inside a C/C++ double quote.
//  Inputs:
//    - token is buffer which holds quote on return
//  Ouputs:
//    - returns 1 if no EOF, otherwise 0
//

int toks::eatDquote(str &token) {
  int ch, esc=0;
  inQuote = 1;
  while((ch = getchr()) != EOF) {
    token.append(ch);
    if ((ch == '\"')&&(!esc)) { // non-escaped double quote
      inQuote = 0; return 1;
    }
    if ((ch == '\\')&&(!esc))   // non-escaped escape char
      esc = 1;
    else
      esc = 0;
  }
  inQuote = 0;
  return 0;
}

//

//----< eat C comment >----------------------------------------
//
//  Processing:
//    - extracts, discards characters inside a C language comment.
//    - works with nested comments
//  Ouputs:
//   	- returns 1 if no EOF, otherwise 0
//

int toks::eatCcom() {
  int ch, chp = 0, level = 1;
  inComm++;
  while((ch = getchr()) != EOF) {
    if ((ch == '/')&&(chp == '*'))
      level--;
    else if (ch == '*') {
      chp = ch;
      ch = getchr();
      if (ch == '/')
	level--;
    }
    if (!level) {
      if(inComm == 1)
        numCL++;   // getChr won't count last comment line
      inComm--;
      return 1;
    }
    chp = ch;
  }
  inComm--;
  return 0;
}

//----< eat C++ comment >--------------------------------------
//
//  Processing:
//    extracts, discards characters inside C++ language comment.
//  Ouputs:
//    returns 1 if no EOF, otherwise 0
//

int toks::eatCPPcom() {
  int ch;
  inComm++;
  while((ch = getchr()) != EOF)
    if (ch == '\n') {
      inComm--;
	  ungetchr(ch);
	  return 1;
    }
  inComm--;
  return 0;
}

//

//---< get token from input stream >---------------------------
//
//  Processing:
//    extracts chars from in and inserts into buffer until
//    non-alphanumeric is encountered
//  Inputs:
//    token  - string provided by client
//  Ouputs:
//    token contains token string on return
//    returns: 0 if end of file else 1
//

int toks::getTok(str &token) {
  char ch;
  token.flush();

// collect contiguous alphnumerics and underscores
  while(isalnum(ch=getchr())||ch=='_')
    token.append(ch);

// if already have a token push back delimiter
  if(token.len()) ungetchr(ch);

// otherwise delimiter is a token, quote, or comment
  else {
    switch(ch) {
      case '\'' : // single quote
        token.append(ch);
        eatSquote(token);
		    return 1;
      case '\"' : // double quote
        token.append(ch);
		    eatDquote(token);
		    return 1;
      case '/'  : // may be a comment
		               ch = getchr();
		    switch(ch) {
		      case '*' : // C comment
			      eatCcom();
			      return getTok(token);
		      case '/' : // C++ comment
			      eatCPPcom();
			      return getTok(token);
		      default  : // it's a delimiter
			      ungetchr(ch);
			      token.append(ch);
		    }
		    break;
      default   : // it's a token
		    token.append(ch);
    }
  }
  if (in.eof())
    return 0;   // end of file
  return 1;     // got a token
}

//

//----< translate token into visible string >------------------
//
//  Processing:
//    - token sent in token str, and is translated in place
//      into a visible string, e.g., an end-of-file character
//      is translated into "<eof>"

void toks::translate(str &token) {

  str tokType(0);
  switch(token[0]) {
    case '\n' : if(token[1] == '\0')
                  tokType = "<newline>";
                break;
    case '\t' : if(token[1] == '\0')
                  tokType = "<tab>";
                break;
    case '\f' : if(token[1] == '\0')
                  tokType = "<formfeed>";
                break;
    case '\r' : if(token[1] == '\0')
                  tokType = "<return>";
                break;
    case ' '  : if(token[1] == '\0')
                  tokType = "<space>";
                break;
    case '\0' :   tokType = "<null token>";
                break;
    case -1   :   tokType = "<eof>";
                break;
    default   : if(((int)token[0] < 32)&&(token[1] == '\0'))
                  tokType = "<cntrl>";
  }
  if(!tokType.len()) return;
  token = tokType;
  return;
}

//

#ifdef TEST_TOKS

#include <fstream.h>     // ifsteam()
#include <iomanip.h>     // setw()
#include <strstrea.h>    // strstream()

//----< test stub >--------------------------------------------

void main(int argc, char* argv[]) {

  #define TOKSIZE  14
  int i,j,notDone;
  str token(TOKSIZE);

  cout << "\n\n   Testing TOKS module \n"
       <<     "  =====================\n\n";

  cout << "   test #1 - tokenizing a string: \n"
       << "  ================================\n\n";

  str s = "The rain in Spain falls mainly on the plain";
  cout << s.cStr() << endl;
  istrstream in((char*)s.cStr());
  toks stoker(in);
  while(stoker.getTok(token))
    if(token.len() > 1)
      cout << token.cStr() << "--";
  cout << "\n\n";

  cout << "   test #2 - tokenizing a file: \n"
       << "  ==============================\n\n";

  if (argc < 2) {
    cout << "please enter file to tokenize on command line\n";
    return;
  }
  else

//

    for (i=1; i<argc; i++) {

    // construct toks object with new stream
      ifstream in(argv[i]);
      if(!in.good()) {
        cout << "  can't open file " << argv[i] << endl;
        continue;
      }
      ftoks toker(in);
      cout << "  tokenizing file " << argv[i] << "\n\n";
      cout << "         tokens : "
              "index of ASCII table for each char\n";
      int save, len;
      do {
        notDone = toker.getTok(token);
        len = token.len();
        save = token[0];
        toker.translate(token);
        cout.setf(ios::right,ios::adjustfield);
        cout << setw(15) << token;
        cout << " : ";
       // send out character indexes
        if(len == 1)
          cout << (int)save;
        else
          for(j=0; token[j]; j++)
            cout << (int)token[j] << " ";
        cout << endl;
       } while(notDone);
       cout << "\n\nend of file\n\n";
      cout << "  #lines         = " << toker.numLines()  << endl;
      cout << "  #comment_lines = " << toker.numCLines() << endl;
      cout << "  #expressions   = " << toker.numExprs()  << endl;
      cout << endl << endl;
      toker.resetStats();
     // construct toks object from existing stream
      const int bufSize = 16;
      str token(bufSize);
      ifstream inAgain(argv[argc-1],ios::in);
      if(inAgain.good()) {
        cout << "first three tokens of last file processed are:"
             << "\n  ";
        ftoks again(inAgain);
        for(i=0; i<3; i++) {
              again.getTok(token);
          again.translate(token);
          cout << " " << token.cStr();
        }
        cout << endl;
      }
    }
}

#endif
